/*===========================================================================
   Textsaic 1.0 - by Santiago Romero aka NoP / Compiler   (c) 2001.
   
      Pixelize is a simple program I wrote to "pixelize" black & white
      and colour images and draw them using text and different colours.
      It's similar to pixelize except for the output, composed of text
      characters instead of plain colour rectangles.
  
      
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 Copyright (c) 2000 Santiago Romero Iglesias
 Email: sromero@escomposlinux.org


 Usage example:

  ./textsaic test.bmp 8 8 textfile [fontsize] > output.html

  (if textfile does not exist, it will use '#' as the output character,
   so you can call textsaic with non-existent files if you want to get
   #-composed images).

   fontsize is an HTML style font size, as +1, +2, -1, -2, 2...

   ./textsaic test.bmp 8 8 abcde -1
  
===========================================================================*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <allegro.h>

/* global variables used */
BITMAP *img;
PALETTE palette;
int window_width=800, window_height=600;
int img_w, img_h;
int square_w, square_h;
int num_squares;
char imagename[512];
char textname[512];
char *text;
                           
/* some useful macros */
#define SCREEN_WIDTH  800 
#define SCREEN_HEIGHT 600
#define RED(pal,col)   (pal[(col)].r)  
#define GREEN(pal,col) (pal[(col)].g)
#define BLUE(pal,col)  (pal[(col)].b)
#define CopyScreen(org,dst) blit((org),(dst),0,0,0,0,(org)->w,(org)->h)

#define HTML_FONT "Courier"
#define COLOR(pal,x) ((pal[x].r*4)<<16 | (pal[(x)].g*4)<<8 | (pal[(x)].b*4) )

/* some function declarations here */
void InitAll( void );
void LoadAll( char * );
void ExitAll( void );
char CharCanBeUsed( char car );
int textsaic( BITMAP *image, int w, int h, 
              int square_width, int square_height, int numcolors );
int displayImagetoText( BITMAP *image, int w, int h,
                        int square_width, int square_height, int numcolors );
void dotextsaic( int x0, int y0, int square_w,
                 int square_h, int numcolors, BITMAP *image );
int NearestColor( int r, int g, int b, PALETTE paleta, int colors );
   
char HTML_SIZE[12];

/*---------------------------------------------------------------------------
 Main function:
  - Read command line arguments.
  - Open image file (.BMP or .PCX) and load it.
  - Do the pixelization process.
  - Write the results on the screen (can be redirected to file).
---------------------------------------------------------------------------*/
int main( int argc, char *argv[] )
{
    int i;
    char car;
    FILE *fp;

    // check arguments
    if( argc < 4 )
    {
      printf("Textsaic Usage: \n");
      printf("\t./textsaic image.bmp square_width square_height [txtfile] [fntsize]\n\n");
      printf("Textsaic will use the # character if no text file is specified.\n");
      printf("Fntsize is an HTML valid font size, as 0, 1, 2, +1 (default), etc.\n");
      exit(-1);
    }

    strcpy( HTML_SIZE, "+1" );
    if( argc > 5 && strlen(argv[5]) < 8 )
       strcpy( HTML_SIZE, argv[5] );
       
    
    // Init Allegro stuff, load the bitmap from disk and setup variables
    InitAll();
    LoadAll(argv[1]);

    // yes, I know the following is pathetic :)
    if( strlen(argv[1]) < 500 ) strcpy( imagename, argv[1] );
    
    img_w = img->w;
    img_h = img->h;
    square_w = atoi(argv[2]);
    square_h = atoi(argv[3]);

    // calculate the number of characters needed and ask for memory for them
    num_squares = ((img_w /square_w)+1)*((img_h / square_h)+1);
    text = malloc( num_squares );
    if( text == NULL )
    {
      fprintf(stderr, "Error allocating memory. Exiting...\n" );
      ExitAll();
    }
    
    // By default use just the '#' character
    for( i=0; i<num_squares; i++ ) text[i] = '#';

    if( argc > 4 )
    {
    
       if( strlen(argv[4]) < 500 ) strcpy( textname, argv[4] );
       
       // Try to read the text file
       fp = fopen( argv[4], "rt" );
       if( fp == NULL )
           fprintf(stderr, 
                   "Error opening file <%s>. Using '#' as character\n",
                   argv[4]);
       else
       {
         for(i=0; i<num_squares; )
         {
            car = fgetc(fp);
            if (car == EOF) i=0;
            if ( CharCanBeUsed(car) ) text[i++] = car;
         }
         fclose(fp);
       }
    }
    
    // do the pixelization process
    readkey(); 
    textsaic( img, img_w, img_h, square_w, square_h, 255 );
    CopyScreen( img, screen );
    readkey();
    
    // display the text output on the screen
    displayImagetoText( img, img_w, img_h, square_w, square_h, 255 );

    if( text != NULL ) free(text);
    ExitAll();
}
END_OF_MAIN();


/*---------------------------------------------------------------------------
 Indicates if char can be used on an HTML text 
---------------------------------------------------------------------------*/
char CharCanBeUsed( char car )
{

    if( car > 32   && car != '\'' &&
        car != '>' && car != '<' &&
        car != '&' && car != '"' )
       return(1);

    return(0);
        
}


/*---------------------------------------------------------------------------
 Textsaic function. It does all the work. 
---------------------------------------------------------------------------*/
int textsaic( BITMAP *image, int w, int h,
              int square_width, int square_height, int numcolors )
{
   int x, y;

   // divide the image in squares and apply doPixelize on them:
   for( y=0; y<h; y+=square_height )
      for( x=0; x<w; x+=square_width )
        dotextsaic( x, y, square_width, square_height, numcolors, image );
    
}



/*---------------------------------------------------------------------------
 Main rendering function: 
   - Read all the pixels on a square and:
      r += pixel_r;     g += pixel_g;        b += pixel_g;
   - Find the most used (r,g,b) resultant colour.
   - Redraw the square using that colour.
---------------------------------------------------------------------------*/
void dotextsaic( int x0, int y0, int square_w, int square_h,
                 int numcolors, BITMAP *image )
{
   int x, y;
   int media = 0, color;
   int media_r = 0, media_g = 0, media_b = 0;
  
   // calculate the media_r, g and b for a given block
   for( y=0; y<square_h; y++ )
   {
      for( x=0; x<square_w; x++ )
      {
         color = getpixel( image, x0+x, y0+y );
         media_r += RED(palette, color);
         media_g += GREEN(palette, color);
         media_b += BLUE(palette, color);
      }
   }
   
   media_r = media_r / (square_w * square_h);
   media_g = media_g / (square_w * square_h);
   media_b = media_b / (square_w * square_h);
   media = NearestColor( media_r, media_g, media_b, palette, 256 );
   
   // repaint all the square using the media
   rectfill( image, x0, y0, x0+square_w, y0+square_h, media );
   
}



/*---------------------------------------------------------------------------
 Show the image as numbers... this reads the image and displays it as
 a HTML file.
---------------------------------------------------------------------------*/
int displayImagetoText( BITMAP *image, int w, int h,
              int square_width, int square_height, 
              int numcolors )
{
   int x, y, z, color, last=666, car;
   PALETTE newpal;

   get_palette( newpal );
   printf("<HTML><BODY BGCOLOR=\"#%X\">", 0 );
   printf("<PRE><FONT FACE=\"%s\" SIZE=\"%s\"><FONT SIZE=\"%s\">", 
           HTML_FONT, HTML_SIZE, HTML_SIZE );
   
   for( z=0, y=0; y<h-square_height; y+=square_height )
   {
      printf("</FONT><BR><FONT SIZE=\"%s\">", HTML_SIZE);
      for( x=0; x<w-square_width; x+=square_width )
      {
         last = color;
         color = getpixel( image, x, y );
         car = text[z++];
         if( color == 0 ) car = ' ';

         if( last != color )
            printf("</FONT><FONT COLOR=\"#%X\">%c",
                     COLOR(newpal,color), car );
         else
            printf("%c", car );
      }  
   }
    
   printf("</FONT></PRE></FONT><P><FONT COLOR=\"#FFFFFF\"><P><BR>");
   printf("Image       : %s\n<BR>", imagename);
   if( strlen(textname) != 0 )
      printf("Text file   : %s\n<BR>", textname);
   printf("Image size  : %d x %d\n<BR>", image->w, image->h );
   printf("Square size : %d x %d\n<BR>", square_width, square_height );
   printf("</FONT></BODY></HTML>\n");
}



/*---------------------------------------------------------------------------
 Allegro Initialization function (inits the graphical system).
---------------------------------------------------------------------------*/
void InitAll( void )
{
   int i;
   
   srand(time(NULL));
   allegro_init();
   install_keyboard();

   set_color_depth(8);
   if (set_gfx_mode(GFX_AUTODETECT, window_width, window_height, 0, 0) != 0)
   {
      set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
      allegro_message("Error setting graphics mode\n%s\n", allegro_error);
      ExitAll();
   }
   
   clear(screen);
}





/*---------------------------------------------------------------------------
 Load function: loads the image from file.
---------------------------------------------------------------------------*/
void LoadAll( char *filename )
{
   img = load_bitmap( filename, palette );
   set_palette( palette );
   CopyScreen( img, screen );
}


/*---------------------------------------------------------------------------
 Exit Function. It exits Pixelize ... :)
---------------------------------------------------------------------------*/
void ExitAll( void )
{
   if( img != NULL ) destroy_bitmap(img);
   set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
   allegro_exit();
   printf("---");
   exit(0);
}


/*---------------------------------------------------------------------------
 int NearestColor( r, g, b, paleta );
 Looks for the nearest (more similar) colour on a palette, giving it
 the (r,g,b) values of the colour to search. It returns -1 on error (it's
 almost impossible to fail).
---------------------------------------------------------------------------*/

int NearestColor( int r, int g, int b, PALETTE paleta, int colors )
{

  int result = -1;       // color mnimo hasta ahora
  int minimo = 200000;   // distancia mnima
  int rd, gd, bd;        // componentes vector c2-c1
  int dist, i;           // distancia e ndice bucle
                                    
  // repeat for N colours:
  for( i=0 ; i<colors ; i++ )
  {
    // calculate vector components as r2-r1, etc
    rd = r-RED(paleta, i);
    gd = g-GREEN(paleta, i);
    bd = b-BLUE(paleta, i);
  
    // calculate colour distances:  
    dist = rd*rd + gd*gd + bd*bd;

    // check if this distance is the smaller till now
    if( dist <minimo )
    {
      // is the lower: annotate and continue.
      minimo = dist ;
      result = i;
    }
  }             
  return( result );
}

